#include "brlcad/cicommon.h"
#include "brlcad/ConstDatabase.h"
#include "displayutil.h"
#include <QImage>
#include <QPixmap>
#include <iostream>


Camera::Camera()
{
	position = new double[3]{-30, 0, 10};
	size = new double[2]{100, 100};
	direction = new double[3]{1, 0, 0};
	rotation = 0.0;
}


// double *
// Camera::getViewportBottomLeftInWorldSpace() {
// 	double *posCoordinates = position.coordinates;
// 	double *dirCoordinates = direction.coordinates;

// 	/* Equation for the camera plane is ax + by + z = A . B, where A = (a,b,c) = direction/normal vector and B = a point on the plane */
// 	double a = dirCoordinates[0];
// 	double b = dirCoordinates[1];
// 	double c = dirCoordinates[2];
// 	double AB = a * posCoordinates[0] + b * posCoordinates[1] + c * posCoordinates[2];

// 	/* Then, if the top left lies at point P, |P| = |size|/2 and position+kw = P, w is the direction pointing towards point P */
// 	BRLCAD::Vector3D pDir = {};
// 	double x = coordinates.x
// 	Vector3D *ret = {}
// 	return 0;
// }


class Callback: public BRLCAD::ConstDatabase::HitCallback {
public:
	double red;
	double blue;
	double green;

	bool operator()(const BRLCAD::ConstDatabase::Hit& hit) throw() {
		if (hit.HasColor()) {
			red = hit.Red();
			green = hit.Green();
			blue = hit.Blue();

			std::cout << red << " " << green << " " << blue << std::endl;
			return true;
		}
		std::cout << "No color for you" << std::endl;
		return false;
	}
};


QPixmap *
DisplayUtils::rayTrace(BRLCAD::ConstDatabase *database, const char *object, Camera *camera)
{
	database->UnSelectAll();
	database->Select(object);
	BRLCAD::Vector3D boxMinima = database->BoundingBoxMinima();
	BRLCAD::Vector3D boxMaxima = database->BoundingBoxMaxima();
	double *min = boxMinima.coordinates;
	double *max = boxMaxima.coordinates;

	double ySize = max[1] - min[1];
	double zSize = max[2] - min[2];

	camera->size[0] = ySize;
	camera->size[1] = zSize;

	/* centre y and z upon object */
	camera->position[1] = (max[1] + min[1])/2;
	camera->position[2] = (max[2] + min[2])/2;
	BRLCAD::Vector3D *vec = new BRLCAD::Vector3D();

	/*
	 * Project rays parallel to the camera's facing direction from the camera's plane onto the object.
	*/
	using namespace BRLCAD;

	double *position = camera->position;
	double *size = camera->size;

	int rows = 500;
	int cols = 500;
	int imageSize = 500;
	int imageHeight;
	int imageWidth;

	if (ySize > zSize) {
		/* wider than tall */
		imageWidth = imageSize;/* /camera->size[0]*camera->size[0]*/
		imageHeight = (int)imageSize/camera->size[0]*camera->size[1];
	} else {
		imageHeight = imageSize;/* /camera->size[1]*camera->size[1]*/
		imageWidth = (int)imageSize/camera->size[1]*camera->size[0];
	}

	rows = std::min(rows, imageHeight);
	cols = std::min(cols, imageWidth);

	/* Evenly distribute all rays across camera viewspace */
	double intervalX = size[0]/(cols - 1);
	double intervalY = size[1]/(rows - 1);

	std::cout << intervalX << " " << intervalY << std::endl;

	/* Start from the bototm left of the camera */
	double *bottomLeft = new double[3]{position[0], position[1] - size[0]/2, position[2] - size[1]/2};

	QImage *image = new QImage(imageWidth, imageHeight, QImage::Format_RGB32);
	image->fill(0);
	image->setPixel(100, 0, QColor(255, 255, 255).rgb());

	for (int i = 0; i < rows; i++) {
		for (int j = 0; j < cols; j++) {
			double rayX = bottomLeft[0];
			double rayY = bottomLeft[1] + intervalX*j;
			double rayZ = bottomLeft[2] + intervalY*i;

			Vector3D *rayOrigin = new Vector3D(rayX, rayY, rayZ);
			Vector3D *rayDirection = new Vector3D(camera->direction);
			Ray3D ray = {*rayOrigin, *rayDirection};
			Callback *callback = new Callback();
			
			// std::cout << "Camera is at: " << camera->position[0] << " " << camera->position[1] << " " << camera->position[2] << std::endl;
			// std::cout << "Shooting from: " << rayX << " " << rayY << " " << rayZ << std::endl;
			// std::cout << "Shooting towards: " << camera->direction[0] << " " << camera->direction[1] << " " << camera->direction[2] << std::endl;

			database->ShootRay(ray, *callback);
			double *location = mapTo2D(rayOrigin, camera);

			int red = (int)(callback->red * 255);
			int green = (int)(callback->green * 255);
			int blue = (int)(callback->blue * 255);

			// std::cout << "RGBint: " << callback->red << " " << callback->green << " " << callback->blue << std::endl;

			int posX = (int)(location[0]*imageWidth);
			int posY = (int)((1-location[1])*imageHeight);

			// std::cout << "2D COORDS:" << posX << " " << posY << std::endl;

			image->setPixel(posX, posY, QColor(red, green, blue).rgb());
		}
	}
	QPixmap *pixmap = new QPixmap();
	pixmap->convertFromImage(*image, (Qt::ImageConversionFlags) Qt::AutoColor);
	return pixmap;
}


/*
 * Return an array with the x and y values, each between 0 and 1, indicative of their location in the viewport.
*/
double *
DisplayUtils::mapTo2D(BRLCAD::Vector3D *vec, Camera *camera)
{
	/* For now, assume that the camera is always facing {1, 0, 0} and rotation = 0*/
	double *vectorCoordinates = vec->coordinates;
	double *cameraCoordinates = camera->position;

	/* X,Y of the 2D plane corresponds to Y,Z of the 3D world */
	double posX = (vectorCoordinates[1] - cameraCoordinates[1])/camera->size[0] + 0.5;
	double posY = (vectorCoordinates[2] - cameraCoordinates[2])/camera->size[1] + 0.5;

	return new double[2]{posX, posY};
}


/*
 * Attempts to map a 2D pixel onto a line in 3D space.
*/
BRLCAD::Vector3D *
DisplayUtils::mapTo3D(double *imageCoordinates, Camera *camera)
{
	/* For now, assume that the camera is always facing {1, 0, 0} and rotation = 0*/
	double *cameraCoordinates = camera->position;

	/* X,Y of the 2D plane corresponds to Y,Z of the 3D world */
	double vecY = (imageCoordinates[0] - 0.5)*camera->size[0] + cameraCoordinates[1];
	double vecZ = (imageCoordinates[1] - 0.5)*camera->size[1] + cameraCoordinates[2];

	return new BRLCAD::Vector3D(0, vecY, vecZ);
}